/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.flink.table.functions.sql

import org.apache.calcite.sql.{SqlFunction, SqlFunctionCategory, SqlKind}
import org.apache.calcite.sql.`type`._

/**
  * All built-in scalar SQL functions.
  */
object ScalarSqlFunctions {

  val E = new SqlFunction(
    "E",
    SqlKind.OTHER_FUNCTION,
    ReturnTypes.DOUBLE,
    null,
    OperandTypes.NILADIC,
    SqlFunctionCategory.NUMERIC)

  val BIN = new SqlFunction(
    "BIN",
    SqlKind.OTHER_FUNCTION,
    ReturnTypes.cascade(ReturnTypes.explicit(SqlTypeName.VARCHAR), SqlTypeTransforms.TO_NULLABLE),
    InferTypes.RETURN_TYPE,
    OperandTypes.family(SqlTypeFamily.INTEGER),
    SqlFunctionCategory.NUMERIC)

  val CONCAT = new SqlFunction(
    "CONCAT",
    SqlKind.OTHER_FUNCTION,
    ReturnTypes.cascade(ReturnTypes.explicit(SqlTypeName.VARCHAR), SqlTypeTransforms.TO_NULLABLE),
    null,
    OperandTypes.ONE_OR_MORE,
    SqlFunctionCategory.STRING)

  val CONCAT_WS = new SqlFunction(
    "CONCAT_WS",
    SqlKind.OTHER_FUNCTION,
    ReturnTypes.cascade(ReturnTypes.explicit(SqlTypeName.VARCHAR), SqlTypeTransforms.TO_NULLABLE),
    null,
    OperandTypes.ONE_OR_MORE,
    SqlFunctionCategory.STRING)

  val LOG = new SqlFunction(
    "LOG",
    SqlKind.OTHER_FUNCTION,
    ReturnTypes.DOUBLE_NULLABLE,
    null,
    OperandTypes.or(OperandTypes.NUMERIC,
      OperandTypes.family(SqlTypeFamily.NUMERIC, SqlTypeFamily.NUMERIC)),
    SqlFunctionCategory.NUMERIC)

  val LPAD = new SqlFunction(
    "LPAD",
    SqlKind.OTHER_FUNCTION,
    ReturnTypes.cascade(
      ReturnTypes.explicit(SqlTypeName.VARCHAR), SqlTypeTransforms.FORCE_NULLABLE),
    null,
    OperandTypes.family(SqlTypeFamily.CHARACTER, SqlTypeFamily.INTEGER, SqlTypeFamily.CHARACTER),
    SqlFunctionCategory.STRING)

  val RPAD = new SqlFunction(
    "RPAD",
    SqlKind.OTHER_FUNCTION,
    ReturnTypes.cascade(
      ReturnTypes.explicit(SqlTypeName.VARCHAR), SqlTypeTransforms.FORCE_NULLABLE),
    null,
    OperandTypes.family(SqlTypeFamily.CHARACTER, SqlTypeFamily.INTEGER, SqlTypeFamily.CHARACTER),
    SqlFunctionCategory.STRING)

  val MD5 = new SqlFunction(
    "MD5",
    SqlKind.OTHER_FUNCTION,
    ReturnTypes.ARG0_NULLABLE,
    InferTypes.RETURN_TYPE,
    OperandTypes.STRING,
    SqlFunctionCategory.STRING
  )

  val SHA1 = new SqlFunction(
    "SHA1",
    SqlKind.OTHER_FUNCTION,
    ReturnTypes.ARG0_NULLABLE,
    InferTypes.RETURN_TYPE,
    OperandTypes.STRING,
    SqlFunctionCategory.STRING
  )

  val SHA256 = new SqlFunction(
    "SHA256",
    SqlKind.OTHER_FUNCTION,
    ReturnTypes.ARG0_NULLABLE,
    InferTypes.RETURN_TYPE,
    OperandTypes.STRING,
    SqlFunctionCategory.STRING
  )

  val DATE_FORMAT = new SqlFunction(
    "DATE_FORMAT",
    SqlKind.OTHER_FUNCTION,
    ReturnTypes.ARG0_NULLABLE,
    InferTypes.RETURN_TYPE,
    OperandTypes.sequence("'(TIMESTAMP, FORMAT)'", OperandTypes.DATETIME, OperandTypes.STRING),
    SqlFunctionCategory.TIMEDATE
  )
}
